"""
utils for nsedt
"""

import io
import json
import datetime
import zipfile

from io import BytesIO

from datetime import datetime, date
from warnings import warn

import requests

import pandas as pd

from fake_http_header import FakeHttpHeader
from nsedt.resources import constants as cns



def format_df(df):
    """
    Arg:\n
        - df: pandas df
    Reuturn:
        - formatted df column 
    """
    df.columns = df.columns.str.lower().str.replace(' ','_').str.replace('\t','')
    return df

def format_date(input_string: str, date_format: str):
    """
    Args:\n
        - input_string : str date format for a format to check
        - format : type of string to format
    Returns:\n
        - str: date format in input string
    """
    try:
        return datetime.strptime(input_string, "%d-%m-%Y").strftime(date_format)
    except ValueError:
        return None


def get_headers():
    """
    Args:
        ---\n
    Returns:\n
        Json: json containing nse header
    """

    return FakeHttpHeader().as_header_dict()


def get_cookies():
    """
    Args:
        ---

    Returns:

        Json: json containing nse cookies
    """

    response = requests.get(cns.BASE_URL, timeout=30, headers=get_headers())
    if response.status_code != 200:
        raise ValueError("Retry again in a minute.")
    return response.cookies.get_dict()


def fetch_url(url, cookies, key=None, response_type="panda_df"):
    """
    Args:

        url (str): URL to fetch
        cookies (str): NSE cookies
        key (str, Optional):

    Returns:

        Pandas DataFrame: df containing url data

    """

    response = requests.get(
        url=url,
        timeout=30,
        headers=get_headers(),
        cookies=cookies,
    )

    if response.status_code == 200:
        json_response = json.loads(response.content)

        if response_type != "panda_df":
            return json_response
        if key is None:
            return pd.DataFrame.from_dict(json_response)

        return pd.DataFrame.from_dict(json_response[key])

    raise ValueError("Please try again in a minute.")


def get_symbol(symbol: str, get_key: str) -> str:
    """_summary_

    Args:
        symbol (str): _description_
        get_key (str): _description_

    Returns:
        str: _description_
    """

    symbol_map = cns.SYMBOL_MAP
    val = None
    for item in symbol_map:
        key_list = item["keys"]
        if symbol in key_list:
            val = item[get_key]

    return val or symbol


def check_nd_convert(start_date: str, end_date: str) -> datetime:
    """
    The function `check_nd_convert` takes two date strings in the format "%d-%m-%Y" and 
    converts them to datetime objects if they are not already in that format.

    :param start_date: The `start_date` parameter is the starting date of a period,
     specified as a string in the format "%d-%m-%Y"
    :type start_date: str
    :param end_date: The `end_date` parameter is a string representing the end date in the format
    "%d-%m-%Y"
    :type end_date: str
    :return: the start_date and end_date as datetime objects.
    """

    if isinstance(start_date, date) and isinstance(end_date, date):
        warn(
            """Passing start_date, end_date in date is deprecated
now pass in str '%d-%m-%Y' format""",
            DeprecationWarning,
            stacklevel=2,
        )

    elif isinstance(start_date, str) and isinstance(end_date, str):
        start_date = datetime.strptime(start_date, "%d-%m-%Y")
        end_date = datetime.strptime(end_date, "%d-%m-%Y")

    else:
        raise ValueError("Input is of an unknown type")

    return start_date, end_date


def fetch_csv(url, cookies, response_type="panda_df", skip_rows=None):
    """
    Args:

        url (str): URL to fetch
        cookies (str): NSE cookies
        key (str, Optional):

    Returns:

        Pandas DataFrame: df generated from csv
        OR
        Json: json output of the csv
        OR
        String: raw content for files where it cannot be processed into Json or 
                Pandas df

    """

    response = requests.get(
        url=url, timeout=30, headers=get_headers(), cookies=cookies )
    if response.status_code == 200:
        if response_type == "raw":
            return response.content
        csv_content = response.content.decode('utf-8')
        df = pd.read_csv(io.StringIO(csv_content), skiprows=skip_rows)
        df = format_df(df)
        return df.to_json(orient='records') if response_type == "json" else df
    raise ValueError("Please try again in a minute.")


def fetch_zip(url, cookies, file_name, response_type="panda_df", skip_rows=None):
    """
    Args:

        url (str): URL to fetch
        cookies (str): NSE cookies
        key (str, Optional):
    Returns:

        Pandas DataFrame: df generated from csv
        OR
        Json: json output of the csv
        OR
        Pandas DF:  Pandas df of the csv file
    """

    if not file_name:
        raise ValueError("Please give file name to return")

    response = requests.get(
        url=url, timeout=30, headers=get_headers(), cookies=cookies )
    if response.status_code == 200:
        zip_content = BytesIO(response.content)
        # Open the zip file in memory
        with zipfile.ZipFile(zip_content, 'r') as zip_ref:
            # Retrieve the list of file names in the zip file
            try:
                csv_content = zip_ref.read(file_name)
            except Exception as e:
                raise ValueError("File not found in the zip folder.") from e

            df = pd.read_csv(BytesIO(csv_content), skiprows=skip_rows)
            df = format_df(df)
            return df.to_json(orient='records') if response_type == "json" else df
    raise ValueError("File might not be available this time or check your params")



"""
function to download reports
"""

import logging

from io import BytesIO
import pandas as pd

from nsedt.utils import get_cookies, fetch_csv, format_date, fetch_zip
from nsedt.resources.constants import (
    REPORT_URL, MARKET_ACTIVITY_REPORT,
    BHAV_COPY_REPORT, SEC_BHAV_COPY_REPORT, NSCCL_REPORTS, NSCCL_VOLT)


log = logging.getLogger("root")

# tbd
# bulk deals
# block deals


def get_market_activity_report(date: str):
    """
    get_market_activity_report

    Args:\n
        date (str): date for which to download market activity report\n
    Returns:
        df = Returns dataframe for  market activity report for sectors
        for the given date
    Expects:
        date to be in format of  "dd-mm-yyyy" eg: 30-04-2024
        all other cases will be invalidated
    """
    date = format_date(date, date_format='%d%m%y')
    if not date:
        raise ValueError("Please provide date format in '%d-%m-%Y' format")

    cookies = get_cookies()
    url = f"{REPORT_URL}{MARKET_ACTIVITY_REPORT}{date}.csv"
    response = fetch_csv(url, cookies, response_type="raw")
    df = pd.read_csv(BytesIO(response), skiprows=8)
    df.drop(columns=['Unnamed: 0'], inplace=True)
    df = df[:77]
    return df

def get_volatility_report(date: str, response_type: str="panda_df"):
    """
    get_volatility_report

    Args:\n
        date (str): date for which to download market activity report\n
        response_type (str, Optional): define the response type panda_df | json . Default json\n
    Returns:
        string: string content of the file as right now its not possible
        to format the content to json or pandas df
    Expects:
        date to be in format of  "dd-mm-yyyy" eg: 30-04-2024
        all other cases will be invalidated
    """
    date = format_date(date, date_format='%d%m%Y')
    if not date:
        raise ValueError("Please provide date format in '%d-%m-%Y' format")
    cookies = get_cookies()
    url = f"{REPORT_URL}{NSCCL_VOLT}CMVOLT_{date}.CSV"
    print(url)
    return fetch_csv(url, cookies, response_type=response_type)


def get_bhav_copy_zip(date: str, response_type: str="panda_df"):
    """
    get_market_activity_report

    Args:\n
        date (str): date for which to download market activity report\n
        path (str): path to save the bhav copy zip
    Returns:
        bool: if the file is save to the local path or not
    Expects:
        date to be in format of  "dd-mm-yyyy" eg: 30-04-2024
    """

    date = format_date(date, date_format='%d%b%Y')
    if not date:
        raise ValueError("Please provide date format in '%d-%m-%Y' format")
    date = date.upper()
    cookies = get_cookies()
    url = f"{REPORT_URL}{BHAV_COPY_REPORT}{date[2:5]}/cm{date}bhav.csv.zip"
    file_name = url.split("/")[-1].replace(".zip", "")
    return fetch_zip(url, cookies, file_name=file_name, response_type=response_type)


def get_sec_full_bhav_copy(date: str, response_type: str="panda_df"):
    """
    get_sec_full_bhav_copy

    Args:\n
        date (str): date for which to download market activity report\n
        response_type (str, Optional): define the response type panda_df | json . Default json\n
    Returns:
        string: string content of the file as right now its not possible
        to format the content to json or pandas df
    Expects:
        date to be in format of  "dd-mm-yyyy" eg: 30-04-2024
        all other cases will be invalidated
    """
    date = format_date(date, date_format='%d%m%Y')
    if not date:
        raise ValueError("Please provide date format in '%d-%m-%Y' format")

    cookies = get_cookies()
    url = f"{REPORT_URL}{SEC_BHAV_COPY_REPORT}sec_bhavdata_full_{date}.csv"
    return fetch_csv(url, cookies, response_type=response_type)


def get_fno_participant_wise_oi_data(date: str, response_type: str="panda_df"):
    """
    fno_participant_wise_oi_data

    Args:\n
        date (str): date for which to download market activity report\n
        response_type (str, Optional): define the response type panda_df | json . Default json\n
    Returns:
        string: string content of the file as right now its not possible
        to format the content to json or pandas df
    Expects:
        date to be in format of  "ddmmYY" eg: 30/04/2024 => 300424
        all other cases will be invalidated
    """
    date = format_date(date, date_format='%d%m%Y')
    if not date:
        raise ValueError("Please provide date format in '%d-%m-%Y' format")

    cookies = get_cookies()
    url = f"{REPORT_URL}{NSCCL_REPORTS}fao_participant_oi_{date}.csv"
    return fetch_csv(url, cookies, response_type=response_type, skip_rows=1)


def get_fno_participant_wise_volume_data(date: str,response_type: str="panda_df"):
    """
    get_fno_participant_wise_volume_data

    Args:\n
        date (str): date for which to download market activity report\n
        response_type (str, Optional): define the response type panda_df | json . Default json\n
    Returns:
        string: string content of the file as right now its not possible
        to format the content to json or pandas df
    Expects:
        date to be in format of  "ddmmYY" eg: 30/04/2024 => 300424
        all other cases will be invalidated
    """
    date = format_date(date, date_format='%d%m%Y')
    if not date:
        raise ValueError("Please provide date format in '%d-%m-%Y' format")

    cookies = get_cookies()
    url = f"{REPORT_URL}{NSCCL_REPORTS}fao_participant_vol_{date}.csv"
    return fetch_csv(url, cookies, response_type=response_type, skip_rows=1)



"""
return data in specific format
"""

import pandas as pd


def price(result):
    """
    Args:
        result (Pandas DataFrame): result

    Returns:
        Pandas DataFrame: df containing data in specific format
    """

    columns_required = [
        "CH_TIMESTAMP",
        "CH_OPENING_PRICE",
        "CH_TRADE_HIGH_PRICE",
        "CH_TRADE_LOW_PRICE",
        "CH_CLOSING_PRICE",
        "CH_PREVIOUS_CLS_PRICE",
        "CH_LAST_TRADED_PRICE",
        "CH_TOT_TRADED_QTY",
        "CH_TOT_TRADED_VAL",
        # "CH_52WEEK_HIGH_PRICE",
        # "CH_52WEEK_LOW_PRICE",
        "VWAP",
        "COP_DELIV_QTY",
        "COP_DELIV_PERC",
        "CH_SERIES",
    ]

    try:
        result = result[columns_required]
    except Exception as e:  # pylint: disable=W0702
        print(f"No data found for the given input: {e}")
        return result
    result = result.set_axis(
        [
            "Date",
            "Open Price",
            "High Price",
            "Low Price",
            "Close Price",
            "Prev Close Price",
            "Last Traded Price",
            "Total Traded Quantity",
            "Total Traded Value",
            # "52 Week High Price",  # Not available in api
            # "52 Week Low Price",  # Not available in api
            "VWAP",
            "Deliverable Volume",
            "Deliverable Percent",
            "Series",
        ],
        axis=1,
    )

    result["Date"] = pd.to_datetime(result["Date"])
    result = result.sort_values("Date", ascending=True)
    result.reset_index(drop=True, inplace=True)
    return result


def indices(
    data_json,
    columns_drop_list: list = None,
    columns_rename_map: map = None,
):
    """
    Args:
        data_json (json):  data in json format
    Returns:
        Pandas DataFrame: df with indexCloseOnlineRecords and indexTurnoverRecords
    """
    if columns_drop_list:
        columns_list = columns_drop_list
    else:
        columns_list = ["_id", "EOD_INDEX_NAME", "TIMESTAMP"]

    if columns_rename_map:
        columns_rename = columns_rename_map
    else:
        columns_rename = {
            "EOD_OPEN_INDEX_VAL": "Open Price",
            "EOD_HIGH_INDEX_VAL": "High Price",
            "EOD_CLOSE_INDEX_VAL": "Close Price",
            "EOD_LOW_INDEX_VAL": "Low Price",
            "EOD_TIMESTAMP": "Date",
        }
    data_close_df = (
        pd.DataFrame(data_json["data"]["indexCloseOnlineRecords"])
        .drop(columns=columns_list)
        .rename(columns=columns_rename)
    )

    ## Mismatch values
    # data_turnover_df = (
    #     pd.DataFrame(d["data"]["indexTurnoverRecords"])
    #     .drop(columns=["_id", "HIT_INDEX_NAME_UPPER", "HIT_TIMESTAMP"])
    #     .rename(
    #         columns={
    #             "HIT_TRADED_QTY": "Total Traded Quantity",
    #             "HIT_TURN_OVER": "Total Traded Value",
    #             "TIMESTAMP": "Date",
    #         }
    #     )
    # )
    return data_close_df
    # return pd.merge(data_close_df, data_turnover_df, on="Date", how="inner")


def option_chain(
    data_json: str,
    response_type: str,
):
    """_summary_

    Args:
        data_json (str): _description_
        response_type (str): _description_

    Returns:
        _type_: _description_
    """
    if response_type == "json":
        data_json_ret = []
        for record in data_json:
            if "PE" in record:
                record["PE"].pop("strikePrice", None)
                record["PE"].pop("expiryDate", None)
                record["PE"].pop("underlying", None)
                record["PE"].pop("identifier", None)
            if "CE" in record:
                record["CE"].pop("strikePrice", None)
                record["CE"].pop("expiryDate", None)
                record["CE"].pop("underlying", None)
                record["CE"].pop("identifier", None)
            data_json_ret.append(record)
        return data_json_ret

    return (
        pd.json_normalize(data_json)
        .sort_values(by=["expiryDate", "strikePrice"], ascending=True)
        .drop(
            columns=[
                "PE.strikePrice",
                "PE.expiryDate",
                "PE.identifier",
                "CE.strikePrice",
                "CE.expiryDate",
                "CE.identifier",
            ]
        )
    )


def get_vix(
    data_json: object,
    response_type: str = "panda_df",
    columns_drop_list: list = None,
):
    """
        Format Vix data
    Args:
        data_json (object): data in json format.
        response_type (str, optional): response_type. Defaults to "panda_df".
        columns_drop_list (list, optional): custom columns drop list. Defaults to None.
    Returns:
        _type_: _description_
    """
    data_json = data_json["data"]
    if columns_drop_list:
        columns_list = columns_drop_list
    else:
        columns_list = [
            "_id",
            "TIMESTAMP",
            "createdAt",
            "updatedAt",
            "__v",
            "ALTERNATE_INDEX_NAME",
            "EOD_INDEX_NAME",
            "EOD_PREV_CLOSE",
            "VIX_PTS_CHG",
            "VIX_PERC_CHG",
        ]
    if response_type == "json":
        data_json_ret = []
        for record in data_json:
            for column in columns_list:
                record.pop(column, None)

            data_json_ret.append(record)
        return data_json_ret

    return (
        pd.json_normalize(data_json)
        .drop(columns=columns_list)
        .rename(
            columns={
                "EOD_OPEN_INDEX_VAL": "Open Price",
                "EOD_HIGH_INDEX_VAL": "High Price",
                "EOD_CLOSE_INDEX_VAL": "Close Price",
                "EOD_LOW_INDEX_VAL": "Low Price",
                "EOD_TIMESTAMP": "Date",
            }
        )
    )


def derivatives_futures(
    data_json: str,
    response_type: str = "panda_df",
    columns_drop_list=None,
):
    """
        Format futures data

    Args:
        data_json (object): data in json format.
        response_type (str, optional): response_type. Defaults to "panda_df".
        columns_drop_list (list, optional): custom columns drop list. Defaults to None.

    Returns:
            json: format data in json
        or
            dataframe: format data in panda df
    """
    if columns_drop_list:
        columns_list = columns_drop_list
    else:
        columns_list = [
            "_id",
            "FH_MARKET_LOT",
            "FH_MARKET_TYPE",
            "FH_OPTION_TYPE",
            "FH_SYMBOL",
            "FH_INSTRUMENT",
            "FH_STRIKE_PRICE",
            "FH_LAST_TRADED_PRICE",
            "TIMESTAMP",
        ]
    if response_type == "json":
        data_json_ret = []
        for record in data_json:
            for column in columns_list:
                record.pop(column, None)
            data_json_ret.append(record)
        return data_json_ret

    return (
        pd.json_normalize(data_json)
        .drop(columns=columns_list)
        .rename(
            columns={
                "FH_OPENING_PRICE": "Open Price",
                "FH_TRADE_HIGH_PRICE": "High Price",
                "FH_CLOSING_PRICE": "Close Price",
                "FH_TRADE_LOW_PRICE": "Low Price",
                "FH_CHANGE_IN_OI": "Change in OI",
                "FH_EXPIRY_DT": "Expiry Date",
                "FH_TIMESTAMP": "Date",
            }
        )
    )


def derivaties_options(
    data_json: str,
    response_type: str = "panda_df",
    columns_drop_list=None,
):
    """
        Format historical options data

    Args:
        data_json (object): data in json format.
        response_type (str, optional): response_type. Defaults to "panda_df".
        columns_drop_list (list, optional): custom columns drop list. Defaults to None.

    Returns:
            json: format data in json
        or
            dataframe: format data in panda df
    """
    if data_json:
        data_json = data_json['data']
    if columns_drop_list:
        columns_list = columns_drop_list
    else:
        columns_list = [
            "_id",
            "FH_MARKET_LOT",
            "FH_MARKET_TYPE",
            "FH_OPTION_TYPE",
            "FH_SYMBOL",
            "FH_INSTRUMENT",
            "FH_STRIKE_PRICE",
            "FH_LAST_TRADED_PRICE",
            "TIMESTAMP",
        ]

    if response_type == "json":
        data_json_ret = []
        for record in data_json:
            for column in columns_list:
                record.pop(column, None)
            data_json_ret.append(record)
        return data_json_ret

    return (
        pd.json_normalize(data_json)
        .drop(columns=columns_list)
        .rename(
            columns={
                "FH_OPENING_PRICE": "Open Price",
                "FH_TRADE_HIGH_PRICE": "High Price",
                "FH_CLOSING_PRICE": "Close Price",
                "FH_TRADE_LOW_PRICE": "Low Price",
                "FH_CHANGE_IN_OI": "Change in OI",
                "FH_EXPIRY_DT": "Expiry Date",
                "FH_TIMESTAMP": "Date",
            }
        )
    )
